"""
Data Inspection Assistant and Visualizations for Polars Dataframe.

Currently, the plot backend is Altair but this is subject to change, and will be decided base on
which plotting backend supports Polars more natively.
"""

from __future__ import annotations

import altair as alt
# import plotly.express as px
# import plotly.graph_objs as go

import polars.selectors as cs
from polars._typing import IntoExpr
from typing import Tuple, List
import polars as pl
import graphviz
import warnings

from typing import Iterable, Dict
from functools import lru_cache
from itertools import combinations
from great_tables import GT

# Internal dependencies
from polars_ds.exprs.ts_features import query_cond_entropy
from polars_ds.exprs.stats import corr
from polars_ds.typing import CorrMethod, PolarsFrame

from .plots import plot_feature_distr

alt.data_transformers.enable("vegafusion")

__all__ = ["DIA"]


# DIA = Data Inspection Assistant / DIAgonsis
class DIA:
    """
    Data Inspection Assistant. Most plots are powered by Altair/great_tables. Altair may require
    additional package downloads.

    If you cannot import this module, please try: pip install "polars_ds[plot]"

    Note: most plots are sampled by default because (1) typically plots don't look good when there
    are too many points, and (2) because of interactivity, if we don't sample, the plots will be too
    large and won't get rendered in a reasonable amount of time. If speed of rendering is crucial and
    you don't need interactivity, use matplotlib.
    """

    # --- Static / Class Methods ---

    # --- Methods ---

    def __init__(self, df: PolarsFrame):
        self._frame: pl.LazyFrame = df.lazy()
        self.numerics: List[str] = df.select(cs.numeric()).collect_schema().names()
        self.ints: List[str] = df.select(cs.integer()).collect_schema().names()
        self.floats: List[str] = df.select(cs.float()).collect_schema().names()
        self.strs: List[str] = df.select(cs.string()).collect_schema().names()
        self.bools: List[str] = df.select(cs.boolean()).collect_schema().names()
        self.cats: List[str] = df.select(cs.categorical()).collect_schema().names()

        schema_dict = df.collect_schema()
        columns = schema_dict.names()

        self.list_floats: List[str] = [
            c
            for c, t in schema_dict.items()
            if (t.is_(pl.List(pl.Float32)) or (t.is_(pl.List(pl.Float64))))
        ]
        self.list_bool: List[str] = [
            c for c, t in schema_dict.items() if t.is_(pl.List(pl.Boolean))
        ]
        self.list_str: List[str] = [c for c, t in schema_dict.items() if t.is_(pl.List(pl.String))]
        self.list_ints: List[str] = [
            c
            for c, t in schema_dict.items()
            if t.is_(pl.List(pl.UInt8))
            or t.is_(pl.List(pl.UInt16))
            or t.is_(pl.List(pl.UInt32))
            or t.is_(pl.List(pl.UInt64))
            or t.is_(pl.List(pl.Int8))
            or t.is_(pl.List(pl.Int16))
            or t.is_(pl.List(pl.Int32))
            or t.is_(pl.List(pl.Int64))
        ]

        self.simple_types: List[str] = (
            self.numerics
            + self.strs
            + self.bools
            + self.cats
            + self.list_floats
            + self.list_ints
            + self.list_bool
            + self.list_str
        )
        self.other_types: List[str] = [c for c in columns if c not in self.simple_types]

    def special_values_report(self) -> pl.DataFrame:
        """
        Checks null, NaN, and non-finite values for float columns. Note that for integers, only null_count
        can possibly be non-zero.
        """
        to_check = self.numerics
        frames = [
            self._frame.select(
                pl.lit(c, dtype=pl.String).alias("column"),
                pl.col(c).null_count().alias("null_count"),
                (pl.col(c).null_count() / pl.len()).alias("null%"),
                pl.col(c).is_nan().sum().alias("NaN_count"),
                (pl.col(c).is_nan().sum() / pl.len()).alias("NaN%"),
                pl.col(c).is_infinite().sum().alias("inf_count"),
                (pl.col(c).is_infinite().sum() / pl.len()).alias("Inf%"),
            )
            for c in to_check
        ]
        return pl.concat(pl.collect_all(frames))

    def numeric_profile(
        self, n_bins: int = 20, iqr_multiplier: float = 1.5, histogram: bool = True, gt: bool = True
    ) -> GT | pl.DataFrame:
        """
        Creates a numerical profile with a histogram plot. Notice that the histograms may have
        completely different scales on the x-axis.

        Parameters
        ----------
        n_bins
            Bins in the histogram
        iqr_multiplier
            Inter Quartile Ranger multiplier. Inter quantile range is the range between
            Q1 and Q3, and this multiplier will enlarge the range by a certain amount and
            use this to count outliers.
        histogram
            Whether to show a histogram or not
        gt
            Whether to show the table as a formatted Great Table or not
        """
        to_check = self.numerics

        cuts = [i / n_bins for i in range(n_bins)]
        cuts[0] -= 1e-5
        cuts[-1] += 1e-5

        if histogram:
            columns_needed = [
                [
                    pl.lit(c, dtype=pl.String).alias("column"),
                    pl.col(c).count().alias("non_null_cnt"),
                    (pl.col(c).null_count() / pl.len()).alias("null%"),
                    pl.col(c).mean().cast(pl.Float64).alias("mean"),
                    pl.col(c).std().cast(pl.Float64).alias("std"),
                    pl.col(c).min().cast(pl.Float64).cast(pl.Float64).alias("min"),
                    pl.col(c).quantile(0.25).cast(pl.Float64).alias("q1"),
                    pl.col(c).median().cast(pl.Float64).round(2).alias("median"),
                    pl.col(c).quantile(0.75).cast(pl.Float64).alias("q3"),
                    pl.col(c).max().cast(pl.Float64).alias("max"),
                    (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25))
                    .cast(pl.Float64)
                    .alias("IQR"),
                    pl.any_horizontal(
                        pl.col(c)
                        < pl.col(c).quantile(0.25)
                        - iqr_multiplier * (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25)),
                        pl.col(c)
                        > pl.col(c).quantile(0.75)
                        + iqr_multiplier * (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25)),
                    )
                    .sum()
                    .alias("outlier_cnt"),
                    pl.struct(
                        ((pl.col(c) - pl.col(c).min()) / (pl.col(c).max() - pl.col(c).min()))
                        .filter(pl.col(c).is_finite())
                        .cut(breaks=cuts, left_closed=True, include_breaks=True)
                        .struct.rename_fields(["brk", "category"])
                        .struct.field("brk")
                        .value_counts()
                        .sort()
                        .struct.field("count")
                        .implode()
                    ).alias("histogram"),
                ]
                for c in to_check
            ]
        else:
            columns_needed = [
                [
                    pl.lit(c, dtype=pl.String).alias("column"),
                    pl.col(c).count().alias("non_null_cnt"),
                    (pl.col(c).null_count() / pl.len()).alias("null%"),
                    pl.col(c).mean().cast(pl.Float64).alias("mean"),
                    pl.col(c).std().cast(pl.Float64).alias("std"),
                    pl.col(c).min().cast(pl.Float64).cast(pl.Float64).alias("min"),
                    pl.col(c).quantile(0.25).cast(pl.Float64).alias("q1"),
                    pl.col(c).median().cast(pl.Float64).round(2).alias("median"),
                    pl.col(c).quantile(0.75).cast(pl.Float64).alias("q3"),
                    pl.col(c).max().cast(pl.Float64).alias("max"),
                    (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25))
                    .cast(pl.Float64)
                    .alias("IQR"),
                    pl.any_horizontal(
                        pl.col(c)
                        < pl.col(c).quantile(0.25)
                        - iqr_multiplier * (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25)),
                        pl.col(c)
                        > pl.col(c).quantile(0.75)
                        + iqr_multiplier * (pl.col(c).quantile(0.75) - pl.col(c).quantile(0.25)),
                    )
                    .sum()
                    .alias("outlier_cnt"),
                ]
                for c in to_check
            ]

        frames = [self._frame.select(*cols) for cols in columns_needed]
        df_final = pl.concat(pl.collect_all(frames))

        if gt:
            gt_out = (
                GT(df_final, rowname_col="column")
                .tab_stubhead("column")
                .fmt_percent(columns="null%")
                .fmt_number(
                    columns=["mean", "std", "min", "q1", "median", "q3", "max", "IQR"], decimals=3
                )
            )
            if histogram:
                return gt_out.fmt_nanoplot(columns="histogram", plot_type="bar")
            return gt_out
        else:
            return df_final

    def col_validation(
        self,
        *rules: Tuple[pl.Expr, str],
    ) -> pl.DataFrame:
        """
        Generates a validation report based on rules (pl.Expr) which evaluates to a single
        boolean per column.

        Parameters
        ----------
        rules
            A tuple of (pl.Expr, str), where the pl.Expr should evaluate to a single boolean value.
            If the boolean is False, then the entire column is considered to be violiating the rule.
        """
        rules_to_check = list(rules)
        rules_exprs = [r.name.keep() for r, _ in rules_to_check]
        violation_messages = [msg for _, msg in rules_to_check]

        df_temp = self._frame.select(*rules_exprs).collect()

        if len(df_temp) > 1:
            raise ValueError(
                "Column rules must evaluate to a single boolean expression for each rule. "
                f"But a dataframe of shape {df_temp.shape} is produced."
            )

        df_violation = df_temp.transpose(include_header=True, column_names=["pass"]).with_columns(
            pl.Series(name="__reason__", values=violation_messages)
        )

        return df_violation.filter(pl.col("pass").not_()).select("column", "__reason__")

    def row_validation(
        self,
        *rules: Tuple[pl.Expr, str],
        id_col: str | None = None,
        columns_to_keep: List[str] | None = None,
        all_reasons: bool = False,
    ) -> pl.DataFrame:
        """
        Generates a validation report based on rules (pl.Expr) which evaluates to booleans
        per row.

        Parameters
        ----------
        rules
            A tuple of (pl.Expr, str), where the pl.Expr should evaluate to a boolean value
            per row. If the boolean is False, then the row is considered a violation. The string
            should be an explanation of the violation.
        id_col
            If None, an "__index__" column will be generated which is the row number.
        columns_to_keep
            Other columns you wish to keep in the final report.
        all_reasons
            If true, all reasons for violations will be returned. If false, only 1 will be returned.
        """

        if id_col is None:
            df = self._frame.with_row_index(name="__index__")
            to_keep = ["__index__"]
        else:
            df = self._frame
            to_keep = [id_col]

        rules_to_check = list(rules)
        rules_exprs = [r.alias(n) for r, n in rules_to_check]
        all_rule_names = [n for _, n in rules_to_check]
        # Do not allow duplicate rule names
        existing_names = set()
        for name in all_rule_names:
            if name not in existing_names:
                existing_names.add(name)
            else:
                raise ValueError(f"Rule name {name} is duplicate. Please rename it.")

        # We cannot use list(set(..)) because that might change the order of all_rule_names

        if columns_to_keep is not None:
            to_keep += columns_to_keep

        df_temp = df.select(*to_keep, *rules_exprs).filter(
            # Filter to the violators.
            # pl.all_horizontal(*all_rule_names) = people who pass all rules
            # pl.all_horizontal(*all_rule_names).not_() = people who failed any one of the rules
            pl.all_horizontal(*all_rule_names).not_()
        )

        if all_reasons:
            reasons = [
                pl.when(pl.col(c)).then(None).otherwise(pl.lit(c, dtype=pl.String))
                for c in all_rule_names
            ]  # When true, return None. When false, return reason

            return df_temp.select(
                *to_keep, pl.concat_list(reasons).list.drop_nulls().list.sort().alias("__reason__")
            ).collect()
        else:
            # df_temp = all people who failed any one of the rules. So there must be at least one 0 in concat-ed list.
            return df_temp.select(
                *to_keep,
                pl.concat_list(all_rule_names)
                .list.arg_min()
                .replace_strict(old=list(range(len(all_rule_names))), new=all_rule_names)
                .alias("__reason__"),
            ).collect()

    def null_corr(
        self,
        subset: IntoExpr | Iterable[IntoExpr] = pl.all(),
        filter_by: pl.Expr | None = None,
    ) -> pl.DataFrame:
        """
        Computes the correlation between A is null and B is null for all (A, B) combinations
        in the given subset of columns.

        If either A or B is all null or all non-null, the null correlation will not be
        computed, since the value is not going to be meaningful.

        Parameters
        ----------
        subset
            Anything that can be put into a Polars .select statement. Defaults to pl.all()
        filter_by
            A boolean expression
        """

        cols = self._frame.select(subset).collect_schema().names()

        if filter_by is None:
            frame = self._frame.select(pl.col(cols).is_null()).collect()
        else:
            frame = self._frame.filter(filter_by).select(pl.col(cols).is_null()).collect()

        df_null_cnt = frame.sum()
        n = frame.shape[0]

        invalid = set(
            c for c, cnt in zip(df_null_cnt.columns, df_null_cnt.row(0)) if (cnt == 0 or cnt == n)
        )

        xx = []
        yy = []
        for x, y in combinations(cols, 2):
            if not (x in invalid or y in invalid):
                xx.append(x)
                yy.append(y)

        if len(xx) == 0:
            return pl.DataFrame(
                {"column_1": [], "column_2": [], "null_corr": []},
                schema={
                    "column_1": pl.String,
                    "column_2": pl.String,
                    "null_corr": pl.Float64,
                },
            )
        else:
            corrs = frame.select(
                pl.corr(x, y).alias(str(i)) for i, (x, y) in enumerate(zip(xx, yy))
            ).row(0)
            return pl.DataFrame({"column_1": xx, "column_2": yy, "null_corr": corrs}).sort(
                pl.col("null_corr").abs(), descending=True
            )

    def meta(self) -> Dict:
        """
        Returns internal data in this class as a dictionary.
        """
        out = self.__dict__.copy()
        out.pop("_frame")
        return out

    def str_stats(self) -> pl.DataFrame:
        """
        Returns basic statistics about the string columns.
        """
        to_check = self.strs
        frames = [
            self._frame.select(
                pl.lit(c).alias("column"),
                pl.col(c).null_count().alias("null_count"),
                pl.col(c).n_unique().alias("n_unique"),
                pl.col(c).value_counts(sort=True).first().struct.field(c).alias("most_freq"),
                pl.col(c)
                .value_counts(sort=True)
                .first()
                .struct.field("count")
                .alias("most_freq_cnt"),
                pl.col(c).str.len_bytes().min().alias("min_byte_len"),
                pl.col(c).str.len_chars().min().alias("min_char_len"),
                pl.col(c).str.len_bytes().mean().alias("avg_byte_len"),
                pl.col(c).str.len_chars().mean().alias("avg_char_len"),
                pl.col(c).str.len_bytes().max().alias("max_byte_len"),
                pl.col(c).str.len_chars().max().alias("max_char_len"),
                pl.col(c).str.len_bytes().quantile(0.05).alias("5p_byte_len"),
                pl.col(c).str.len_bytes().quantile(0.95).alias("95p_byte_len"),
            )
            for c in to_check
        ]
        return pl.concat(pl.collect_all(frames))

    def corr(
        self, subset: IntoExpr | Iterable[IntoExpr], method: CorrMethod = "pearson"
    ) -> pl.DataFrame:
        """
        Returns a dataframe containing correlation information between the subset and all numeric columns.
        Only numerical columns will be checked.

        Parameters
        ----------
        subset
            Anything that can be put into a Polars .select statement.
        method
            One of ["pearson", "spearman", "xi", "kendall", "bicor"]
        """

        to_check = self._frame.select(subset).collect_schema().names()

        corrs = [
            self._frame.select(
                # This calls corr from .stats
                pl.lit(x).alias("column"),
                *(corr(x, y, method=method).alias(y) for y in self.numerics),
            )
            for x in to_check
        ]

        return pl.concat(pl.collect_all(corrs))

    def plot_corr(
        self, subset: IntoExpr | Iterable[IntoExpr], method: CorrMethod = "pearson"
    ) -> GT:
        """
        Plots the correlations using classic heat maps.

        Parameters
        ----------
        subset
            Anything that can be put into a Polars .select statement.
        method
            One of ["pearson", "spearman", "xi", "kendall", "bicor"]
        """
        corr_values = self.corr(subset, method)
        cols = [c for c in corr_values.columns if c != "column"]
        return (
            GT(corr_values)
            .fmt_number(columns=cols, decimals=3)
            .data_color(
                columns=cols,
                palette=["#0202bd", "#bd0237"],
                domain=[-1, 1],
                alpha=0.5,
                na_color="#000000",
            )
        )

    def infer_prob(self) -> List[str]:
        """
        Infers columns that can potentially be probabilities. For f32/f64 columns, this checks if all values are
        between 0 and 1. For List[f32] or List[f64] columns, this checks whether the column can potentially be
        multi-class probabilities.
        """
        is_ok = (
            self._frame.select(
                *((pl.col(c).is_between(0.0, 1.0).all()).alias(c) for c in self.floats),
                *(
                    (
                        (
                            pl.col(c).list.eval((pl.element() >= 0.0).all()).list.first()
                        )  # every number must be positive
                        & ((pl.col(c).list.sum() - 1.0).abs() < 1e-6)  # class prob must sum to 1
                        & (
                            pl.col(c).list.len().min() == pl.col(c).list.len().max()
                        )  # class prob column must have the same length
                    ).alias(c)
                    for c in self.list_floats
                ),
            )
            .collect()
            .row(0)
        )

        return [c for c, ok in zip(self.floats + self.list_floats, is_ok) if ok is True]

    @lru_cache
    def infer_high_null(self, threshold: float = 0.75) -> List[str]:
        """
        Infers columns with more than threshold percentage nulls.

        Parameters
        ----------
        threshold
            The threshold above which a column will be considered high null
        """
        is_ok = (
            self._frame.select(
                (pl.col(c).null_count() >= pl.len() * threshold).alias(c)
                for c in self._frame.columns
            )
            .collect()
            .row(0)
        )

        return [c for c, ok in zip(self._frame.columns, is_ok) if ok is True]

    @lru_cache
    def infer_discrete(self, threshold: float = 0.1, max_val_cnt: int = 100) -> List[str]:
        """
        Infers discrete columns based on unique percentage and max_val_count.

        Parameters
        ----------
        threshold
            Columns with unique percentage lower than threshold will be considered
            discrete
        max_val_cnt
            Max number of unique values the column can have in order for it to be considered
            discrete
        """
        out: List[str] = self.bools + self.cats
        to_check = [c for c in self._frame.columns if c not in out]
        is_ok = (
            self._frame.select(
                (
                    (pl.col(c).n_unique() < max_val_cnt)
                    | (pl.col(c).n_unique() < threshold * pl.len())
                ).alias(c)
                for c in to_check
            )
            .collect()
            .row(0)
        )

        return [c for c, ok in zip(to_check, is_ok) if ok is True]

    @lru_cache
    def infer_const(self, include_null: bool = False) -> List[str]:
        """
        Infers whether the column is constant.

        Parameters
        ----------
        include_null
            If true, a constant column with null values will also be included.
        """
        if include_null:
            is_ok = (
                self._frame.select(
                    (
                        (pl.col(c).n_unique() == 1)
                        | ((pl.col(c).null_count() > 0) & (pl.col(c).n_unique() == 2))
                    ).alias(c)
                    for c in self._frame.columns
                )
                .collect()
                .row(0)
            )
        else:
            is_ok = (
                self._frame.select(
                    (pl.col(c).n_unique() == 1).alias(c) for c in self._frame.columns
                )
                .collect()
                .row(0)
            )

        return [c for c, ok in zip(self._frame.columns, is_ok) if ok is True]

    @lru_cache
    def infer_binary(self, include_null: bool = False) -> List[str]:
        """
        Infers whether the column is binary.

        Parameters
        ----------
        include_null
            If true, a binary column with 2 non-null distinct values and null will also be included.
        """
        if include_null:
            is_ok = (
                self._frame.select(
                    (
                        (pl.col(c).n_unique() == 2)
                        | ((pl.col(c).null_count() > 0) & (pl.col(c).n_unique() == 3))
                    ).alias(c)
                    for c in self._frame.columns
                )
                .collect()
                .row(0)
            )
        else:
            is_ok = (
                self._frame.select(
                    (pl.col(c).n_unique() == 2).alias(c) for c in self._frame.columns
                )
                .collect()
                .row(0)
            )

        return [c for c, ok in zip(self._frame.columns, is_ok) if ok is True]

    @lru_cache
    def infer_k_distinct(self, k: int, include_null: bool = False) -> List[str]:
        """
        Infers whether the column has k distinct values.

        Parameters
        ----------
        k
            Any positive integer.
        include_null
            If true, a binary column with k non-null distinct values and null will also be included.
        """
        if k < 1:
            raise ValueError("Input `k` must be >= 1.")

        if include_null:
            is_ok = (
                self._frame.select(
                    (
                        (pl.col(c).n_unique() == k)
                        | ((pl.col(c).null_count() > 0) & (pl.col(c).n_unique() == (k + 1)))
                    ).alias(c)
                    for c in self._frame.columns
                )
                .collect()
                .row(0)
            )
        else:
            is_ok = (
                self._frame.select(
                    (pl.col(c).n_unique() == k).alias(c) for c in self._frame.columns
                )
                .collect()
                .row(0)
            )

        return [c for c, ok in zip(self._frame.columns, is_ok) if ok is True]

    def infer_corr(self, method: CorrMethod = "pearson") -> pl.DataFrame:
        """
        Trying to infer highly correlated columns by computing correlation between
        all numerical (including boolean) columns.

        Parameters
        ----------
        method
            One of ["pearson", "spearman", "xi", "kendall"]
        """
        to_check = self.numerics + self.bools

        xx = []
        yy = []
        for x, y in combinations(to_check, 2):
            xx.append(x)
            yy.append(y)

        corrs = (
            self._frame.with_columns(pl.col(c).cast(pl.UInt8) for c in self.bools)
            .select(corr(x, y, method=method).alias(f"{i}") for i, (x, y) in enumerate(zip(xx, yy)))
            .collect()
            .row(0)
        )

        return pl.DataFrame({"x": xx, "y": yy, "corr": corrs}).sort(
            pl.col("corr").abs(), descending=True
        )

    def infer_dependency(self, subset: IntoExpr | Iterable[IntoExpr] = pl.all()) -> pl.DataFrame:
        """
        Infers (functional) dependency using the method of conditional entropy. This only evaluates
        potential qualifying columns. Potential qualifying columns are columns of type:
        int, str, categorical, or booleans.

        If returned conditional entropy is very low, that means knowning the column in
        `by` is enough to to infer the column in `column`, or the column in `column` can
        be determined by the column in `by`.

        Parameters
        ----------
        subset
            A subset of columns to try running the dependency check. The subset input can be
            anything that can be turned into a Polars selector. The df or the column subset of the df
            may contain columns that cannot be used for dependency detection, e.g. column of list of values.
            Only valid columns will be checked.
        """

        # Infer valid columns to run this detection
        valid = self.ints + self.strs + self.cats + self.bools
        check_frame = self._frame.select(subset)
        all_names = check_frame.collect_schema().names()
        to_check = [x for x in all_names if x in valid]
        n_uniques = check_frame.select(pl.col(c).n_unique() for c in to_check).collect().row(0)

        frame = (
            pl.DataFrame({"column": to_check, "n_unique": n_uniques})
            .filter(pl.col("n_unique") > 1)
            .sort("n_unique")
        )

        check = list(frame["column"])
        if len(check) <= 1:
            warnings.warn(
                f"Not enough valid columns to detect dependency on. Valid column count: {len(check)}. Empty dataframe returned.",
                stacklevel=2,
            )
            return pl.DataFrame(
                {"column": [], "by": [], "cond_entropy": []},
                schema={"column": pl.String, "by": pl.String, "cond_entropy": pl.Float64},
            )

        if len(check) != len(all_names):
            warnings.warn(
                f"The following columns are dropped because they cannot be used in dependency detection: {[f for f in all_names if f not in check]}",
                stacklevel=2,
            )

        # Construct output
        column = []
        by = []
        for x, y in combinations(check, 2):
            column.append(x)
            by.append(y)

        ce = (
            self._frame.select(
                query_cond_entropy(x, y).abs().alias(f"{i}")
                for i, (x, y) in enumerate(zip(column, by))
            )
            .collect()
            .row(0)
        )

        out = pl.DataFrame({"column": column, "by": by, "cond_entropy": ce}).sort("cond_entropy")

        return out

    def plot_dependency(
        self, threshold: float = 0.01, subset: IntoExpr | Iterable[IntoExpr] = pl.all()
    ) -> graphviz.Digraph:
        """
        Plot dependency using the result of self.infer_dependency and positively dtermines
        dependency by the threshold.

        Parameters
        ----------
        threshold
            If conditional entropy is < threshold, we draw a line indicating dependency.
        subset
            A subset of columns to try running the dependency check. The subset input can be
            anything that can be turned into a Polars selector
        """

        dep_frame = self.infer_dependency(subset=subset)

        df_local = dep_frame.filter((pl.col("cond_entropy") < threshold)).select(
            pl.col("column").alias("child"),  # c for child
            pl.col("by").alias("parent"),  # p for parent
        )
        cp = df_local.group_by("child").agg(pl.col("parent"))
        pc = df_local.group_by("parent").agg(pl.col("child"))
        child_parent: dict[str, pl.Series] = dict(zip(cp["child"], cp["parent"]))
        parent_child: dict[str, pl.Series] = dict(zip(pc["parent"], pc["child"]))

        dot = graphviz.Digraph(
            "Dependency Plot", comment=f"Conditional Entropy < {threshold:.2f}", format="png"
        )
        for c, par in child_parent.items():
            parents_of_c = set(par)
            for p in par:
                # Does parent p have a child that is also a parent of c? If so, remove p.
                children_of_p = parent_child.get(p, None)
                if children_of_p is not None:
                    if len(parents_of_c.intersection(children_of_p)) > 0:
                        parents_of_c.remove(p)

            dot.node(c)
            for p in parents_of_c:
                dot.node(p)
                dot.edge(p, c)

        return dot

    def plot_feature_distr(
        self,
        feature: str,
        n_bins: int | None = None,
        density: bool = False,
        show_bad_values: bool = True,
        min_: float | pl.Expr | None = None,
        max_: float | pl.Expr | None = None,
        over: str | None = None,
        filter_by: pl.Expr | None = None,
    ) -> alt.Chart:
        """
        Plot distribution of the feature with a few statistical details.

        Parameters
        ----------
        feature
            A string representing a column name
        n_bins
            The number of bins used for histograms. Not used when the feature column is categorical.
        density
            Whether to plot a probability density or not
        show_bad_values
            Whether to show % of bad (null or non-finite) values
        min_
            Whether to ignore values strictly lower than min_
        max_
            Whether to ignore values strictly higher than max_
        over
            Whether to look at the distribution over another categorical column
        filter_by
            An extra condition you may want to impose on the underlying dataset
        """
        if feature not in self.numerics:
            raise ValueError("Input feature must be numeric.")

        conditions = []
        if filter_by is not None:
            conditions.append(filter_by)
        if min_ is not None:
            conditions.append(pl.col(feature) >= min_)
        if max_ is not None:
            conditions.append(pl.col(feature) <= max_)

        if len(conditions) > 0:
            df = self._frame.filter(pl.all_horizontal(*conditions)).select(feature, over).collect()
        else:
            df = self._frame.select(feature, over).collect()

        return plot_feature_distr(
            df=df,
            feature=feature,
            n_bins=n_bins,
            density=density,
            show_bad_values=show_bad_values,
            over=over,
        )



from __future__ import annotations

import numpy as np
import polars as pl
import polars_ds as pds
import pytest
import sklearn.metrics

SEED = 208
N_ROWS = 2_000

np.random.seed(SEED)


Y_TRUE = np.random.choice([True, False], N_ROWS)
Y_SCORE = np.random.rand(N_ROWS)
Y_PRED = Y_SCORE > 0.5
Y_TRUE_SC = np.full(N_ROWS, True)
Y_SCORE_SC = np.ones(N_ROWS)
Y_PRED_SC = Y_SCORE_SC > 0.5

# The point of the combos is to test our code under a matrix of different scenarios:
# any combination of all classes present in y_true, only one class present in y_true,
# a range of scores present in y_score, only one score present in y_score, all classes
# present in y_pred, and only one class present in y_pred.

TRUE_PRED_COMBOS = [
    (Y_TRUE, Y_PRED),
    (Y_TRUE_SC, Y_PRED),
    (Y_TRUE, Y_PRED_SC),
    (Y_TRUE_SC, Y_PRED_SC),
]

TRUE_SCORE_COMBOS = [
    (Y_TRUE, Y_SCORE),
    (Y_TRUE, Y_SCORE_SC),
    (Y_TRUE_SC, Y_SCORE),
    (Y_TRUE_SC, Y_SCORE_SC),
]


def nandiv(a: np.float64, b: np.float64) -> np.float64:
    if b == 0:
        return np.nan

    return a / b


def reference_confusion_matrix(y_true, y_pred):
    tn, fp, fn_, tp = sklearn.metrics.confusion_matrix(
        y_true, y_pred, labels=[False, True]
    ).ravel()  # Setting labels will ensure that the matrix is zero-filled

    # Where possible, test against sklearn. Otherwise implement it by hand.
    p = tp + fn_
    n = fp + tn
    tpr = nandiv(tp, p)
    fnr = 1.0 - tpr
    fpr = nandiv(fp, n)
    tnr = 1.0 - fpr
    precision = sklearn.metrics.precision_score(
        y_true, y_pred, labels=[False, True], zero_division=np.nan
    )
    false_omission_rate = nandiv(fn_, fn_ + tn)
    plr = nandiv(tpr, fpr)
    nlr = nandiv(fnr, tnr)
    npv = 1.0 - false_omission_rate
    fdr = 1.0 - precision
    prevalence = nandiv(p, p + n)
    informedness = tpr + tnr - 1.0
    prevalence_threshold = nandiv(np.sqrt(tpr * fpr) - fpr, tpr - fpr)
    markedness = precision - false_omission_rate
    dor = nandiv(plr, nlr)
    balanced_accuracy = sklearn.metrics.balanced_accuracy_score(y_true, y_pred)
    f1 = sklearn.metrics.f1_score(y_true, y_pred, zero_division=np.nan)
    folkes_mallows_index = sklearn.metrics.fowlkes_mallows_score(y_true, y_pred)
    mcc = sklearn.metrics.matthews_corrcoef(y_true, y_pred)
    acc = nandiv(tp + tn, p + n)
    threat_score = nandiv(tp, tp + fn_ + fp)

    return (
        tn,
        fp,
        fn_,
        tp,
        tpr,
        fpr,
        fnr,
        tnr,
        prevalence,
        prevalence_threshold,
        informedness,
        precision,
        false_omission_rate,
        plr,
        nlr,
        acc,
        balanced_accuracy,
        f1,
        folkes_mallows_index,
        mcc,
        threat_score,
        markedness,
        fdr,
        npv,
        dor,
    )


@pytest.mark.parametrize("y_true,y_score", TRUE_PRED_COMBOS)
def test_confusion_matrix(y_true, y_score):
    ref = reference_confusion_matrix(y_true, y_score > 0.5)
    res = [
        x.to_list()[0]
        for x in pl.DataFrame({"y_true": y_true, "y_score": y_score})
        .select(pds.query_confusion_matrix("y_true", "y_score", all_metrics=True).alias("metrics"))
        .unnest("metrics")
        .iter_columns()
    ]

    pytest.approx(res) == ref


def test_roc_auc():
    from sklearn.metrics import roc_auc_score

    df = pds.frame(size=2000).select(
        pds.random(0.0, 1.0).alias("predictions"),
        pds.random(0.0, 1.0).round().cast(pl.Int32).alias("target"),
        pl.lit(0).alias("zero_target"),
    )

    roc_auc = df.select(pds.query_roc_auc("target", "predictions")).item(0, 0)

    answer = roc_auc_score(df["target"].to_numpy(), df["predictions"].to_numpy())

    assert np.isclose(roc_auc, answer)

    # When all classes are 0, roc_auc returns NaN
    nan_roc = df.select(pds.query_roc_auc("zero_target", "predictions")).item(0, 0)

    assert np.isnan(nan_roc)

    # This is an edge case where we have only one value for predicted prob.
    # This is just guessing at random so 0.5 is the right output.
    # Technical reason:
    # (TPR, FPR becomes 1 value, which messes up the trapz calculation, and we need to fill
    # a 0 at the beginning to make sure the trapz calculation is always valid)
    df = pl.DataFrame({"a": [0, 1], "b": [0.5, 0.5]})
    result = df.select(pds.query_roc_auc(pl.col("a"), pl.col("b"))).item(0, 0)
    assert result == 0.5


def test_roc_auc_2():
    from sklearn.metrics import roc_auc_score

    # A test submitted by a user. PDS didn't have 0 padding for TPR and FPR before the user
    # submitted the issue and this test is added to check that. This behavior is consistent
    # with scipy's trapz calculation and therefore with sklearn's roc auc.
    df = pl.from_dict(
        {
            "ytrue": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            "ypred": [1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1],
        }
    )

    roc_auc_pds = df.select(pds.query_roc_auc("ytrue", "ypred")).item(0, 0)

    roc_auc_sklearn = roc_auc_score(df["ytrue"].to_numpy(), df["ypred"].to_numpy())

    assert np.isclose(roc_auc_pds, roc_auc_sklearn)


def test_multiclass_roc_auc():
    from sklearn.metrics import roc_auc_score

    def roc_auc_random_data(size: int = N_ROWS) -> pl.DataFrame:
        df = pds.frame(size=N_ROWS, index_name="id").with_columns(
            pl.col("id").cast(pl.UInt64),
            pds.random().alias("val1"),
            pds.random().alias("val2"),
            pds.random().alias("val3"),
            pl.col("id").mod(3).alias("actuals"),
        )
        # Need to normalize to make sure this is valid ROC AUC data
        return (
            df.lazy()
            .with_columns(
                pl.concat_list(
                    pl.col("val1")
                    / pl.sum_horizontal(pl.col("val1"), pl.col("val2"), pl.col("val3")),
                    pl.col("val2")
                    / pl.sum_horizontal(pl.col("val1"), pl.col("val2"), pl.col("val3")),
                    pl.col("val3")
                    / pl.sum_horizontal(pl.col("val1"), pl.col("val2"), pl.col("val3")),
                ).alias("pred")
            )
            .select(
                pl.col("actuals"),
                pl.col("pred"),
            )
            .collect()
        )

    df = roc_auc_random_data()
    y_pred = np.stack(df["pred"].to_numpy())
    y_true = df["actuals"]

    macro = df.select(pds.query_multi_roc_auc("actuals", "pred", 3, "macro")).item(0, 0)
    macro_sklearn = roc_auc_score(y_true, y_pred, average="macro", multi_class="ovr")

    weighted = df.select(pds.query_multi_roc_auc("actuals", "pred", 3, "weighted")).item(0, 0)
    weighted_sklearn = roc_auc_score(y_true, y_pred, average="weighted", multi_class="ovr")

    assert np.isclose(macro, macro_sklearn, rtol=1e-10, atol=1e-12)
    assert np.isclose(weighted, weighted_sklearn, rtol=1e-10, atol=1e-10)


def test_precision_recall_roc_auc():
    import numpy as np
    from sklearn.metrics import roc_auc_score

    df = pl.DataFrame(
        {
            "a": np.random.random(size=N_ROWS),
            "b": np.random.random(size=N_ROWS),
            "y": np.round(np.random.random(size=N_ROWS)).astype(int),
        }
    )
    for threshold in [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]:
        res = df.select(
            pds.query_binary_metrics("y", "a", threshold=threshold).alias("metrics")
        ).unnest("metrics")
        precision_res = res.get_column("precision")[0]
        recall_res = res.get_column("recall")[0]
        roc_auc_res = res.get_column("roc_auc")[0]

        # precision, recall by hand
        predicted_prob = np.array(df["a"])
        predicted = predicted_prob >= threshold  # boolean
        actual = np.array(df["y"])  # .to_numpy()
        precision = actual[predicted].sum() / np.sum(predicted)
        recall = ((actual == 1) & (predicted == 1)).sum() / (actual.sum())

        assert np.isclose(precision, precision_res)
        assert np.isclose(recall, recall_res)
        assert np.isclose(roc_auc_score(actual, predicted_prob), roc_auc_res)


def test_log_loss():
    import numpy as np
    from sklearn.metrics import log_loss

    df = pl.DataFrame(
        {
            "x": np.random.random(size=N_ROWS),
            "y": np.round(np.random.random(size=N_ROWS)).astype(int),
        }
    )

    res = df.select(pds.query_log_loss("y", "x")).item(0, 0)
    ans = log_loss(df["y"].to_numpy(), df["x"].to_numpy())
    assert np.isclose(res, ans, rtol=1e-10)


def test_dcg_score():
    from sklearn.metrics import dcg_score

    df = pds.frame(size=1000, index_name="id").with_columns(
        pl.col("id").mod(3).alias("example_id"),
        pds.random_int(0, 10).alias("relevance"),
        pds.random().alias("score"),
    )

    for tie in [True, False]:
        df_agg = (
            df.group_by("example_id")
            .agg(dcg_score=pds.query_dcg_score("relevance", "score", ignore_ties=tie))
            .sort("example_id")
        )

        sklearn_results = []
        for by, subdf in df.sort("example_id").partition_by("example_id", as_dict=True).items():
            rel = subdf["relevance"].to_numpy().reshape((1, -1))
            score = subdf["score"].to_numpy().reshape((1, -1))
            sklearn_results.append(dcg_score(rel, score, ignore_ties=tie))

        sklearn_results = np.array(sklearn_results)

        np.testing.assert_allclose(df_agg["dcg_score"].to_numpy(), sklearn_results, rtol=1e-5)

    for k in range(1, 5):
        df_agg = (
            df.group_by("example_id")
            .agg(dcg_score=pds.query_dcg_score("relevance", "score", k=k, ignore_ties=False))
            .sort("example_id")
        )

        sklearn_results = []
        for by, subdf in df.sort("example_id").partition_by("example_id", as_dict=True).items():
            rel = subdf["relevance"].to_numpy().reshape((1, -1))
            score = subdf["score"].to_numpy().reshape((1, -1))
            sklearn_results.append(dcg_score(rel, score, k=k, ignore_ties=False))

        sklearn_results = np.array(sklearn_results)

        np.testing.assert_allclose(df_agg["dcg_score"].to_numpy(), sklearn_results, rtol=1e-5)


def test_ndcg_score():
    from sklearn.metrics import ndcg_score

    df = pds.frame(size=1000, index_name="id").with_columns(
        pl.col("id").mod(3).alias("example_id"),
        pds.random_int(0, 10).alias("relevance"),
        pds.random().alias("score"),
    )

    for tie in [True, False]:
        df_agg = (
            df.group_by("example_id")
            .agg(ndcg_score=pds.query_ndcg_score("relevance", "score", ignore_ties=tie))
            .sort("example_id")
        )

        sklearn_results = []
        for by, subdf in df.sort("example_id").partition_by("example_id", as_dict=True).items():
            rel = subdf["relevance"].to_numpy().reshape((1, -1))
            score = subdf["score"].to_numpy().reshape((1, -1))
            sklearn_results.append(ndcg_score(rel, score, ignore_ties=tie))

        sklearn_results = np.array(sklearn_results)
        np.testing.assert_allclose(df_agg["ndcg_score"].to_numpy(), sklearn_results, rtol=1e-5)

    for k in range(1, 5):
        df_agg = (
            df.group_by("example_id")
            .agg(ndcg_score=pds.query_ndcg_score("relevance", "score", k=k, ignore_ties=False))
            .sort("example_id")
        )

        sklearn_results = []
        for by, subdf in df.sort("example_id").partition_by("example_id", as_dict=True).items():
            rel = subdf["relevance"].to_numpy().reshape((1, -1))
            score = subdf["score"].to_numpy().reshape((1, -1))
            sklearn_results.append(ndcg_score(rel, score, k=k, ignore_ties=False))

        sklearn_results = np.array(sklearn_results)
        np.testing.assert_allclose(df_agg["ndcg_score"].to_numpy(), sklearn_results, rtol=1e-5)



"""Miscallaneous Numerical Functions and Transforms."""

from __future__ import annotations
import math
import polars as pl
from typing import List, Iterable, TYPE_CHECKING

# Internal dependencies
from polars_ds.typing import (
    DetrendMethod,
    ConvMode,
    ConvMethod,
)
from polars_ds._utils import pl_plugin, to_expr

if TYPE_CHECKING:
    from numpy import ndarray

__all__ = [
    "singular_values",
    "principal_components",
    "pca",
    "softmax",
    "gcd",
    "lcm",
    "haversine",
    "jaccard_row",
    "jaccard_col",
    "psi",
    "psi_w_breakpoints",
    "psi_discrete",
    "woe",
    "woe_discrete",
    "info_value",
    "info_value_discrete",
    "integrate_trapz",
    "convolve",
    "list_amax",
    "gamma",
    "expit",
    "exp2",
    "expit",
    "logit",
    "trunc",
    "detrend",
    "rfft",
    "fract",
    "center",
    "z_normalize",
    "isotonic_regression",
    "is_increasing",
    "is_decreasing",
    "next_up",
    "next_down",
    "digamma",
    "xlogy",
    "add_at",
    "l_inf_horizontal",
    "l1_horizontal",
    "l2_sq_horizontal",
    # "mutual_info_disc",
]


def l_inf_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
    """
    Horizontally computes L inf norm. Shorthand for pl.max_horizontal(pl.col(x).abs() for x in exprs).

    Parameters
    ----------
    *v
        Expressions to compute horizontal L infinity.
    normalize
        Whether to divide by the dimension
    """
    if normalize:
        exprs = list(v)
        return pl.max_horizontal(to_expr(x).abs() for x in exprs) / len(exprs)
    else:
        return pl.max_horizontal(to_expr(x).abs() for x in v)


def l2_sq_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
    """
    Horizontally computes L2 norm squared. Shorthand for pl.sum_horizontal(pl.col(x).pow(2) for x in exprs).

    Parameters
    ----------
    *v
        Expressions to compute horizontal L2.
    normalize
        Whether to divide by the dimension
    """
    if normalize:
        exprs = list(v)
        return pl.sum_horizontal(to_expr(x).pow(2) for x in exprs) / len(exprs)
    else:
        return pl.sum_horizontal(to_expr(x).pow(2) for x in v)


def l1_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
    """
    Horizontally computes L1 norm. Shorthand for pl.sum_horizontal(pl.col(x).abs() for x in exprs).

    Parameters
    ----------
    *v
        Expressions to compute horizontal L1.
    normalize
        Whether to divide by the dimension
    """
    if normalize:
        exprs = list(v)
        return pl.sum_horizontal(to_expr(x).abs() for x in exprs) / len(exprs)
    else:
        return pl.sum_horizontal(to_expr(x).abs() for x in v)


def is_increasing(x: str | pl.Expr, strict: bool = False) -> pl.Expr:
    """
    Checks whether the column is monotonically increasing.

    Parameters
    ----------
    x
        A numerical column
    strict
        Whether the check should be strict
    """
    if strict:
        return (to_expr(x).diff() > 0.0).all()
    else:
        return (to_expr(x).diff() >= 0.0).all()


def is_decreasing(x: str | pl.Expr, strict: bool = False) -> pl.Expr:
    """
    Checks whether the column is monotonically decreasing.

    Parameters
    ----------
    x
        A numerical column
    strict
        Whether the check should be strict
    """
    xx = to_expr(x)
    if strict:
        return (xx.diff() < 0.0).all()
    else:
        return (xx.diff() <= 0.0).all()


def center(x: str | pl.Expr) -> pl.Expr:
    """
    Centers the column.

    This is only a short cut for a standard feature transform, and is not recommended
    to be used in settings where the means need to be persisted.
    """
    xx = to_expr(x)
    return xx - xx.mean()


def z_normalize(x: str | pl.Expr) -> pl.Expr:
    """
    Z-normalizes the column.

    This is only a short cut for a standard feature transform, and is not recommended
    to be used in settings where the means/stds need to be persisted.
    """
    xx = to_expr(x)
    mean = xx.mean()
    std = xx.std()
    return (xx - mean) / std


def softmax(x: str | pl.Expr) -> pl.Expr:
    """
    Applies the softmax function to the column, which turns any real valued column into valid probability
    values. This is simply a shorthand for x.exp() / x.exp().sum() for expressions x.

    Paramters
    ---------
    x
        Either a str represeting a column name or a Polars expression
    """
    xx = to_expr(x)
    return xx.exp() / (xx.exp().sum())


def gcd(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
    """
    Computes GCD of two integer columns. This will try to cast everything to int32.

    Parameters
    ----------
    x
        An integer column
    y
        Either an int, or another integer column
    """
    if isinstance(y, int):
        yy = pl.lit(y, dtype=pl.Int32)
    else:
        yy = to_expr(y).cast(pl.Int32)

    return pl_plugin(
        symbol="pl_gcd",
        args=[to_expr(x).cast(pl.Int32), yy],
        is_elementwise=True,
    )


def lcm(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
    """
    Computes LCM of two integer columns. This will try to cast everything to int32.

    Parameters
    ----------
    x
        An integer column
    y
        Either an int, or another integer column
    """
    if isinstance(y, int):
        yy = pl.lit(y, dtype=pl.Int32)
    else:
        yy = to_expr(y).cast(pl.Int32)

    return pl_plugin(
        symbol="pl_lcm",
        args=[to_expr(x).cast(pl.Int32), yy],
        is_elementwise=True,
    )


def haversine(
    x_lat: str | pl.Expr,
    x_long: str | pl.Expr,
    y_lat: float | str | pl.Expr,
    y_long: float | str | pl.Expr,
) -> pl.Expr:
    """
    Computes haversine distance using the naive method. The output unit is km.

    Parameters
    ----------
    x_lat
        Column representing latitude in x
    x_long
        Column representing longitude in x
    y_lat
        Column representing latitude in y
    y_long
        Column representing longitude in y
    """
    xlat = to_expr(x_lat)
    xlong = to_expr(x_long)
    ylat = pl.lit(y_lat) if isinstance(y_lat, float) else to_expr(y_lat)
    ylong = pl.lit(y_long) if isinstance(y_long, float) else to_expr(y_long)
    return pl_plugin(
        symbol="pl_haversine",
        args=[xlat, xlong, ylat, ylong],
        is_elementwise=True,
        cast_to_supertype=True,
    )


def singular_values(
    *features: str | pl.Expr,
    center: bool = True,
    as_explained_var: bool = False,
    as_ratio: bool = False,
) -> pl.Expr:
    """
    Finds all principal values (singular values) for the data matrix formed by the given features
    and returns them in descending order.

    Note: if a row has null values, it will be dropped.

    Paramters
    ---------
    features
        Feature columns
    center
        Whether to center the data or not. If you want to standard-normalize, set this to False,
        and do it for input features by hand.
    as_explained_var
        If true, return the explained variance, which is singular_value ^ 2 / (n_samples - 1)
    as_ratio
        If true, normalize output to between 0 and 1.
    """
    feats = [to_expr(f) for f in features]
    if center:
        actual_inputs = [f - f.mean() for f in feats]
    else:
        actual_inputs = feats

    out = pl_plugin(symbol="pl_singular_values", args=actual_inputs, returns_scalar=True)
    if as_explained_var:
        out = out.list.eval(pl.element().pow(2) / (pl.count() - 1))
    if as_ratio:
        out = out.list.eval(pl.element() / pl.element().sum())

    return out


def pca(
    *features: str | pl.Expr,
    center: bool = True,
) -> pl.Expr:
    """
    Finds all singular values as well as the principle vectors.

    Paramters
    ---------
    features
        Feature columns
    center
        Whether to center the data or not. If you want to standard normalize, set this to False,
        and do it for input features by hand.
    """
    feats = [to_expr(f) for f in features]
    if center:
        actual_inputs = [f - f.mean() for f in feats]
    else:
        actual_inputs = feats

    return pl_plugin(
        symbol="pl_pca", args=actual_inputs, changes_length=True, pass_name_to_apply=True
    )


def principal_components(
    *features: str | pl.Expr,
    k: int = 2,
    center: bool = True,
) -> pl.Expr:
    """
    Transforms the features to get the first k principal components. This returns NaN if the number
    of rows is less than `k`.

    Paramters
    ---------
    features
        Feature columns
    center
        Whether to center the data or not. If you want to standard normalize, set this to False,
        and do it for input features by hand.
    """
    feats = [to_expr(f) for f in features]
    if k > len(feats) or k <= 0:
        raise ValueError("Input `k` should be between 1 and the number of features inclusive.")

    actual_inputs = [pl.lit(k, dtype=pl.UInt32)]
    if center:
        actual_inputs.extend(f - f.mean() for f in feats)
    else:
        actual_inputs.extend(feats)

    return pl_plugin(symbol="pl_principal_components", args=actual_inputs, pass_name_to_apply=True)


def jaccard_row(a: str | pl.Expr, b: str | pl.Expr) -> pl.Expr:
    """
    Computes jaccard similarity pairwise between `a` and `b` column. The type of
    each column must be list and the lists must have the same inner type. The inner type
    must either be integer or string.

    Parameters
    ----------
    a
        A list column with a hashable inner type
    b
        A list column with a hashable inner type
    """
    aa = to_expr(a)
    bb = to_expr(b)
    intersection_len = aa.list.set_intersection(bb).list.len()
    a_len = aa.list.len()
    b_len = bb.list.len()
    return intersection_len / (a_len + b_len - intersection_len)


def jaccard_col(a: str | pl.Expr, b: str | pl.Expr, count_null: bool = False) -> pl.Expr:
    """
    Computes jaccard similarity column-wise. This will hash entire columns and compares the two
    hashsets. Note: only integer/str columns can be compared.

    Parameters
    ----------
    a
        A column with a hashable type
    b
        A column with a hashable type
    count_null
        Whether to count null as a distinct element.
    """
    aa = to_expr(a).unique()
    bb = to_expr(b).unique()

    if not count_null:
        aa = aa.drop_nulls()
        bb = bb.drop_nulls()

    return jaccard_row(aa.implode(), bb.implode())


def psi(
    new: str | pl.Expr | Iterable[float],
    baseline: str | pl.Expr | Iterable[float],
    n_bins: int = 10,
    return_report: bool = False,
) -> pl.Expr:
    """
    Compute the Population Stability Index between x and the reference column (usually x's historical values).
    The reference column will be divided into n_bins quantile bins which will be used as basis of comparison.

    Note this assumes values in self and ref are continuous. This will also remove all infinite, null, NA.
    values.

    Also note that it will try to create `n_bins` many unique breakpoints. If input data has < n_bins
    unique breakpoints, the repeated breakpoints will be grouped together, and the computation will be done
    with < `n_bins` many bins. This happens when a single value appears too many times in data. This also
    differs from the reference implementation by treating breakpoints as right-closed intervals with -inf
    and inf being the first and last values of the intervals. This is because we need to accommodate all data
    in the case when actual data's min and the reference data's min are not the same, which is common in reality.

    Parameters
    ----------
    new
        An expression or any iterable that can be turned into a Polars series that represents newly
        arrived feature values
    baseline
        An expression or any iterable that can be turned into a Polars series. Usually this should
        be the feature's historical values
    n_bins : int, > 1
        The number of quantile bins to use
    return_report
        Whether to return a PSI report or not.

    Reference
    ---------
    https://github.com/mwburke/population-stability-index/blob/master/psi.py
    https://www.listendata.com/2015/05/population-stability-index.html
    """
    if n_bins <= 1:
        raise ValueError("Input `n_bins` must be >= 2.")

    if isinstance(new, (str, pl.Expr)):
        new_ = to_expr(new)
        valid_new = new_.filter(new_.is_finite()).cast(pl.Float64)
    else:
        temp = pl.Series(values=new, dtype=pl.Float64)
        valid_new = pl.lit(temp.filter(temp.is_finite()))

    if isinstance(baseline, (str, pl.Expr)):
        base = to_expr(baseline)
        valid_ref = base.filter(base.is_finite()).cast(pl.Float64)
    else:
        temp = pl.lit(pl.Series(values=baseline, dtype=pl.Float64))
        valid_ref = temp.filter(temp.is_finite())

    vc = (
        valid_ref.qcut(n_bins, left_closed=False, allow_duplicates=True, include_breaks=True)
        .struct.rename_fields(
            ["brk", "category"]
        )  # Use "breakpoints" in the future. Skip this rename. After polars v1
        .struct.field("brk")
        .value_counts()
        .sort()
    )

    # breakpoints learned from ref
    brk = vc.struct.field("brk")  # .cast(pl.Float64)
    # counts of points in the buckets
    cnt_ref = vc.struct.field("count")  # .cast(pl.UInt32)
    psi_report = pl_plugin(
        symbol="pl_psi_report",
        args=[valid_new, brk, cnt_ref],
        changes_length=True,
    ).alias("psi_report")
    if return_report:
        return psi_report

    return psi_report.struct.field("psi_bin").sum()


def psi_discrete(
    new: str | pl.Expr | Iterable[float],
    baseline: str | pl.Expr | Iterable[float],
    return_report: bool = False,
) -> pl.Expr:
    """
    Compute the Population Stability Index between self (actual) and the reference column. The baseline
    column will be used as categories which are the basis of comparison.

    Note this assumes values in new and ref baseline discrete columns (e.g. str categories). This will
    treat each value as a distinct category and null will be treated as a category by itself. If a category
    exists in new but not in baseline, the percentage will be imputed by 0.0001. If you do not wish to include
    new distinct values in PSI calculation, you can still compute the PSI by generating the report and filtering.

    Also note that discrete columns must have the same type in order to be considered the same.

    Parameters
    ----------
    new
        The feature
    baseline
        An expression, or any iterable that can be turned into a Polars series. Usually this should
        be the historical values
    return_report
        Whether to return a PSI report or not.

    Reference
    ---------
    https://www.listendata.com/2015/05/population-stability-index.html
    """
    if isinstance(new, (str, pl.Expr)):
        new_ = to_expr(new)
        temp = new_.value_counts().struct.rename_fields(["", "count"])
        new_cnt = temp.struct.field("count")
        new_cat = temp.struct.field("")
    else:
        temp = pl.Series(values=new)
        temp = temp.value_counts()  # This is a df in this case
        new_cnt = pl.lit(temp.drop_in_place("count"))
        new_cat = pl.lit(temp[temp.columns[0]])

    if isinstance(baseline, (str, pl.Expr)):
        base = to_expr(baseline)
        temp = base.value_counts().struct.rename_fields(["", "count"])
        ref_cnt = temp.struct.field("count")
        ref_cat = temp.struct.field("")
    else:
        temp = pl.Series(values=baseline)
        temp = temp.value_counts()  # This is a df in this case
        ref_cnt = pl.lit(temp.drop_in_place("count"))
        ref_cat = pl.lit(temp[temp.columns[0]])

    psi_report = pl_plugin(
        symbol="pl_psi_discrete_report",
        args=[new_cat, new_cnt, ref_cat, ref_cnt],
        changes_length=True,
    )
    if return_report:
        return psi_report

    return psi_report.struct.field("psi_bin").sum()


def psi_w_breakpoints(
    new: str | pl.expr | Iterable[float],
    baseline: str | pl.expr | Iterable[float],
    breakpoints: List[float],
) -> pl.Expr:
    """
    Creates a PSI report using the custom breakpoints.

    Parameters
    ----------
    new
        The data representing the new observed data. Any sequence of numerical values that
        can be turned into a Polars'series, or an expression representing a column will work
    baseline
        The data representing the baseline data. Any sequence of numerical values that
        can be turned into a Polars'series, or an expression representing a column will work
    breakpoints
        The data that represents breakpoints. Input must be sorted, distinct, finite numeric values.
        This function will not cleanse the breakpoints for the user. E.g. [0.1, 0.5, 0.9] will create
        four bins: (-inf. 0.1], (0.1, 0.5], (0.5, 0.9] and (0.9, inf). Please do not pass inf or NaN values
        as breakpoints.
    """
    if isinstance(baseline, (str, pl.Expr)):
        x: pl.Expr = to_expr(baseline)
        x = x.filter(x.is_finite())
    else:
        temp = pl.Series(values=baseline)
        x: pl.Expr = pl.lit(temp.filter(temp.is_finite()))

    if isinstance(new, (str, pl.Expr)):
        y: pl.Expr = to_expr(new)
        y = y.filter(y.is_finite())
    else:
        temp = pl.Series(values=new)
        y: pl.Expr = pl.lit(temp.filter(temp.is_finite()))

    if len(breakpoints) == 0:
        raise ValueError("Breakpoints is empty.")

    bp = breakpoints + [float("inf")]
    return pl_plugin(
        symbol="pl_psi_w_bps",
        args=[x.rechunk(), y.rechunk(), pl.Series(values=bp)],
        changes_length=True,
    ).alias("psi_report")


def woe(x: str | pl.Expr, target: str | pl.expr | Iterable[float], n_bins: int = 10) -> pl.Expr:
    """
    Compute the Weight of Evidence for x with respect to target. This assumes x
    is continuous. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation.

    Currently only quantile binning strategy is implemented.

    Parameters
    ----------
    x
        The feature
    target
        The target variable. Should be 0s and 1s.
    n_bins
        The number of bins to bin the variable.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    if isinstance(target, (str, pl.Expr)):
        t = to_expr(target)
    else:
        t = pl.Series(values=target)
    xx = to_expr(x)
    valid = xx.filter(xx.is_finite())
    brk = valid.qcut(n_bins, left_closed=False, allow_duplicates=True).cast(pl.String)
    return pl_plugin(symbol="pl_woe_discrete", args=[brk, t], changes_length=True)


def woe_discrete(
    x: str | pl.Expr,
    target: str | pl.Expr | Iterable[int],
) -> pl.Expr:
    """
    Compute the Weight of Evidence for x with respect to target. This assumes x
    is discrete and castable to String. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation.

    Parameters
    ----------
    x
        The feature
    target
        The target variable. Should be 0s and 1s.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    if isinstance(target, (str, pl.Expr)):
        t = to_expr(target)
    else:
        t = pl.Series(values=target)
    return pl_plugin(
        symbol="pl_woe_discrete",
        args=[to_expr(x).cast(pl.String), t],
        changes_length=True,
    )


def info_value(
    x: str | pl.Expr,
    target: str | pl.expr | Iterable[float],
    n_bins: int = 10,
    return_sum: bool = True,
) -> pl.Expr:
    """
    Compute Information Value for x with respect to target. This assumes the variable x
    is continuous. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation.

    Currently only quantile binning strategy is implemented.

    Parameters
    ----------
    x
        The feature. Must be numeric.
    target
        The target column. Should be 0s and 1s.
    n_bins
        The number of bins to bin x.
    return_sum
        If false, the output is a struct containing the ranges and the corresponding IVs. If true,
        it is the sum of the individual information values.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    if isinstance(target, (str, pl.Expr)):
        t = to_expr(target)
    else:
        t = pl.Series(values=target)
    xx = to_expr(x)
    valid = xx.filter(xx.is_finite())
    brk = valid.qcut(n_bins, left_closed=False, allow_duplicates=True).cast(pl.String)
    out = pl_plugin(symbol="pl_iv", args=[brk, t], changes_length=True)
    return out.struct.field("iv").sum() if return_sum else out


def info_value_discrete(
    x: str | pl.Expr, target: str | pl.Expr | Iterable[int], return_sum: bool = True
) -> pl.Expr:
    """
    Compute the Information Value for x with respect to target. This assumes x
    is discrete and castable to String. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation.

    Parameters
    ----------
    x
        The feature. The column must be castable to String
    target
        The target variable. Should be 0s and 1s.
    return_sum
        If false, the output is a struct containing the categories and the corresponding IVs. If true,
        it is the sum of the individual information values.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    if isinstance(target, (str, pl.Expr)):
        t = to_expr(target)
    else:
        t = pl.Series(values=target)
    out = pl_plugin(symbol="pl_iv", args=[to_expr(x).cast(pl.String), t], changes_length=True)
    return out.struct.field("iv").sum() if return_sum else out


def integrate_trapz(y: str | pl.Expr, x: float | pl.Expr) -> pl.Expr:
    """
    Integrate y along x using the trapezoidal rule. If x is not a single
    value, then x should be sorted.

    Parameters
    ----------
    y
        A column of numbers
    x
        If it is a single float, it must be positive and it will represent a uniform
        distance between points. If it is an expression, it must be sorted, does not contain
        null, and have the same length as self.
    """
    yy = to_expr(y).cast(pl.Float64).rechunk()
    if isinstance(x, float):
        xx = pl.lit(abs(x), pl.Float64)
    else:
        xx = to_expr(x).cast(pl.Float64)

    return pl_plugin(
        symbol="pl_trapz",
        args=[yy, xx],
        returns_scalar=True,
    )


def convolve(
    x: str | pl.Expr,
    kernel: List[float] | ndarray | pl.Series | pl.Expr,  # noqa: F821
    fill_value: float | pl.Expr = 0.0,
    method: ConvMethod = "direct",
    mode: ConvMode = "full",
    parallel: bool = False,
) -> pl.Expr:
    """
    Performs a convolution with the given kernel(filter). The current implementation's performance is worse
    than SciPy but offers parallelization within Polars.

    For large kernels (usually kernel length > 120), convolving with FFT is faster, but for smaller kernels,
    convolving with direct method is faster.

    parameters
    ----------
    x
        A column of numbers
    kernel
        The filter for the convolution. Anything that can be turned into a Polars Series will work. All non-finite
        values will be filtered out before the convolution.
    fill_value
        Fill null values in `x` with this value. Either a float or a polars's expression representing 1 element
    method
        Either `fft` or `direct`.
    mode
        Please check the reference. One of `same`, `left` (left-aligned same), `right` (right-aligned same),
        `valid` or `full`.
    parallel
        Only applies when method is `direct`. Whether to compute the convulotion in parallel. Note that this may not
        have the expected performance when you are in group_by or other parallel context already. It is recommended
        to use this in select/with_columns context, when few expressions are being run at the same time.

    Reference
    ---------
    https://brianmcfee.net/dstbook-site/content/ch03-convolution/Modes.html
    https://en.wikipedia.org/wiki/Convolution
    """
    xx = to_expr(x).fill_null(fill_value).cast(pl.Float64).rechunk()  # One cont slice
    f: pl.Expr | pl.Series
    if isinstance(kernel, pl.Expr):
        f = kernel.filter(kernel.is_finite()).rechunk()  # One cont slice
    else:
        f = pl.Series(values=kernel, dtype=pl.Float64)
        f = f.filter(f.is_finite()).rechunk()  # One cont slice

    if method == "direct":
        f = f.reverse()

    return pl_plugin(
        symbol="pl_convolve",
        args=[xx, f],
        kwargs={"mode": mode, "method": method, "parallel": parallel},
        changes_length=True,
    )


def list_amax(list_col: str | pl.Expr) -> pl.Expr:
    """
    Finds the argmax of the list in this column. This is useful for

    (1) Turning sparse multiclass target into dense target.
    (2) Finding the max probability class of a multiclass classification output.
    (3) As a shortcut for expr.list.eval(pl.element().arg_max()).
    """
    return to_expr(list_col).list.eval(pl.element().arg_max())


def gamma(x: str | pl.Expr) -> pl.Expr:
    """
    Applies the gamma function to self. Note, this will return NaN for negative values and inf when x = 0,
    whereas SciPy's gamma function will return inf for all x <= 0.
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_gamma",
        is_elementwise=True,
    )


def expit(x: str | pl.Expr) -> pl.Expr:
    """
    Applies the Expit function to self. Expit(x) = 1 / (1 + e^(-x))
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_expit",
        is_elementwise=True,
    )


def logit(x: str | pl.Expr) -> pl.Expr:
    """
    Applies the logit function to self. Logit(x) = ln(x/(1-x)).
    Note that logit(0) = -inf, logit(1) = inf, and logit(p) for p < 0 or p > 1 yields nan.
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_logit",
        is_elementwise=True,
    )


def exp2(x: str | pl.Expr) -> pl.Expr:
    """
    Returns 2^x.
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_exp2",
        is_elementwise=True,
    )


def fract(x: str | pl.Expr) -> pl.Expr:
    """
    Returns the fractional part of the input values. E.g. fractional part of 1.1 is 0.1
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_fract",
        is_elementwise=True,
    )


def trunc(x: str | pl.Expr) -> pl.Expr:
    """
    Returns the integer part of the input values. E.g. integer part of 1.1 is 1.0
    """
    return pl_plugin(
        args=[to_expr(x)],
        symbol="pl_trunc",
        is_elementwise=True,
    )


def sinc(x: str | pl.Expr) -> pl.Expr:
    """
    Computes the sinc function normalized by pi.
    """
    xx = to_expr(x)
    y = math.pi * pl.when(xx == 0).then(1e-20).otherwise(xx)
    return y.sin() / y


def xlogy(x: str | pl.Expr, y: str | pl.Expr) -> pl.Expr:
    """
    Computes x * log(y) so that if x = 0, the product is 0.

    Parameters
    ----------
    x
        A numerical column
    y
        A numerical column
    """
    return pl_plugin(
        args=[to_expr(x).cast(pl.Float64), to_expr(y).cast(pl.Float64)],
        symbol="pl_xlogy",
        is_elementwise=True,
    )


def detrend(x: str | pl.Expr, method: DetrendMethod = "linear") -> pl.Expr:
    """
    Detrends self using either linear/mean method. This does not persist.

    Parameters
    ----------
    method
        Either `linear` or `mean`
    """
    ts = to_expr(x)
    if method == "linear":
        N = ts.count()
        x = pl.int_range(0, N, eager=False)
        coeff = pl.cov(ts, x) / x.var()
        const = ts.mean() - coeff * (N - 1) / 2
        return ts - x * coeff - const
    elif method == "mean":
        return ts - ts.mean()
    else:
        raise ValueError(f"Unknown detrend method: {method}")


def rfft(series: str | pl.Expr, n: int | None = None, return_full: bool = False) -> pl.Expr:
    """
    Computes the DFT transform of a real-valued input series using FFT Algorithm. Note that
    by default a series of length (length // 2 + 1) will be returned.

    Parameters
    ----------
    series
        Input real series
    n
        The number of points to use. If n is smaller than the length of the input,
        the input is cropped. If it is larger, the input is padded with zeros.
        If n is not given, the length of the input is used.
    return_full
        If true, output will have the same length as determined by n.
    """
    if n is not None and n <= 1:
        raise ValueError("Input `n` should be > 1.")

    full = pl.lit(return_full, pl.Boolean)
    nn = pl.lit(n, pl.UInt32)
    x: pl.Expr = to_expr(series).cast(pl.Float64)
    return pl_plugin(symbol="pl_rfft", args=[x, nn, full], changes_length=True)


def target_encode(
    s: str | pl.Expr,
    target: str | pl.Expr | Iterable[int],
    min_samples_leaf: int = 20,
    smoothing: float = 10.0,
) -> pl.Expr:
    """
    Compute information necessary to target encode a string column.

    Note: nulls will be encoded as well.

    Parameters
    ----------
    s
        The string column to encode
    target
        The target column. Should be 0s and 1s.
    min_samples_leaf
        A regularization factor
    smoothing
        Smoothing effect to balance categorical average vs prior

    Reference
    ---------
    https://contrib.scikit-learn.org/category_encoders/targetencoder.html
    """
    if isinstance(target, (str, pl.Expr)):
        t = to_expr(target)
    else:
        t = pl.lit(pl.Series(values=target))
    return pl_plugin(
        symbol="pl_target_encode",
        args=[to_expr(s), t, t.mean()],
        kwargs={"min_samples_leaf": float(min_samples_leaf), "smoothing": smoothing},
        changes_length=True,
    )


def isotonic_regression(
    y: str | pl.Expr, weights: str | pl.Expr | None = None, increasing: bool = True
) -> pl.Expr:
    """
    Performs isotonic regression on the data. This is the same as scipy.optimize.isotonic_regression.

    Parameters
    ----------
    y
        The response variable
    weights
        The weights for the response
    increasing
        If true, output will be monotonically inreasing. If false, it will be monotonically
        decreasing.
    """

    yy = to_expr(y).cast(pl.Float64)
    args = [yy]
    has_weights = weights is not None
    if has_weights:
        args.append(to_expr(weights).cast(pl.Float64))

    return pl_plugin(
        symbol="pl_isotonic_regression",
        args=args,
        kwargs={
            "has_weights": has_weights,
            "increasing": increasing,
        },
    )


def next_up(x: str | pl.Expr) -> pl.Expr:
    """
    For any float, return the least number greater than itself (within the precision).
    Intergers will be treated as f32. E.g. The next value up for 0.1 is 0.10000000000000002
    because of precision issues. This is useful when you need to make extremely small changes
    to certain values and you don't want to add random noise.
    """
    return pl_plugin(
        symbol="pl_next_up",
        args=[to_expr(x)],
        is_elementwise=True,
    )


def next_down(x: str | pl.Expr) -> pl.Expr:
    """
    For any float, return the greatest number smaller than itself (within the precision).
    Intergers will be treated as f32. E.g. The next value down for 0.1 is 0.09999999999999999.
    This is useful when you need to make extremely small changes to certain values and you don't
    want to add random noise.
    """
    return pl_plugin(
        symbol="pl_next_down",
        args=[to_expr(x)],
        is_elementwise=True,
    )


def digamma(x: str | pl.Expr) -> pl.Expr:
    """
    The diagamma function.
    """
    return pl_plugin(
        symbol="pl_diagamma",
        args=[to_expr(x)],
        is_elementwise=True,
    )


def add_at(
    indices: str | pl.Expr, values: str | pl.Expr, buffer_size: int | pl.Expr | None = None
) -> pl.Expr:
    """
    Creates a zero column of length `buffer_size` first. Then the j-th value in `values`
    will be added to the j-th index in `indices` in the buffer. This is the equivalent to
    NumPy's add.at.

    Parameters
    ----------
    indices
        Expression or name of a column. Must be castable to u32.
    values
        Expression or name of a column. Must be castable to f64 and have the same length
        as the indices.
    buffer_size
        If this is None, buffer size will be inferred from unique values in indices, which
        should range from [0..n), where n is the actual buffer size. If this is an integer,
        then the buffer will have the exact size given here, which might cause out of bounds
        error if indices are not checked. If this is an expression, only the first element
        in the represented column will be used.
    """
    ind = to_expr(indices).cast(pl.UInt32).rechunk()
    val = to_expr(values).cast(pl.Float64).rechunk()
    if isinstance(buffer_size, pl.Expr):
        size = buffer_size
    else:
        size = pl.lit(buffer_size, dtype=pl.UInt32)

    return pl_plugin(
        symbol="pl_add_at",
        args=[ind, val, size],
    )


# Not sure what is wrong here

# def mutual_info_disc(
#     x: str | pl.Expr,
#     target: str | pl.Expr,
#     n_neighbors: int = 3,
#     seed: int | None = None,
# ) -> float:
#     """
#     Computes the mutual infomation between a continuous variable x and a discrete
#     target varaible. Note: (1) This always assume `x` is continuous. (2) Unlike Scikit-learn,
#     if a target category has <= `n_neighbors` records, then the result here may not make any sense.
#     This is because (a) `n_neighbors` is typically a small number and if your target category has so
#     few records, then maybe you need to rethink the target definition, and (b) if we assume target category
#     always has > `n_neighbors` records, then we can speed up the algorithm by quite a bit.


#     Reference
#     ---------
#     https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0087357
#     """

#     if n_neighbors <= 0:
#         raise ValueError("Input `n_neighbors` must be > 0.")

#     feat = to_expr(x).cast(pl.Float64)
#     feat = feat / feat.std(ddof=0)
#     t = to_expr(target)

#     scale_factor = pl.max_horizontal(pl.lit(1.0, dtype=pl.Float64), feat.abs().mean())

#     c = pl_plugin(
#         symbol="pl_jitter",
#         args=[feat, 1e-10 * scale_factor, pl.lit(seed, dtype=pl.UInt64)],
#         is_elementwise=True,
#     )

#     kwargs = {
#         "k": n_neighbors,
#         "metric": "l1",
#         "parallel": False,
#         "skip_eval": False,
#         "max_bound": 99999.0,
#         "epsilon": 0.
#     }

#     # This is not really exposed to the user. It does `dist_from_kth_nb` and a `next_down` in one go.
#     r = pl_plugin(
#         symbol="pl_dist_from_kth_nb",
#         args=[c],
#         kwargs=kwargs,
#     ).over(t)

#     label_counts = c.len().over(t)

#     # NB Cnt contains the point itself. This is what we want
#     # This corresponds to m_i in the paper.
#     nb_cnt = pl_plugin(
#         symbol="pl_nb_cnt",
#         args=[r, c],
#         kwargs={
#             "k": 0,
#             "metric": "l1",  # Data is 1d, l2 metric is equal to l1, but l2 does more instructions. So use l1
#             "parallel": False,
#             "skip_eval": False,
#             "skip_data": False,
#         },
#     )


#     # psi in SciPy is the diagamma function
#     psi_label_counts = pl_plugin(
#         symbol="pl_diagamma",
#         args=[label_counts],
#         is_elementwise=True,
#     )

#     psi_nb_cnt = pl_plugin(
#         symbol="pl_diagamma",
#         args=[nb_cnt],
#         is_elementwise=True,
#     )

#     psi_n_samples = pl_plugin(
#         symbol="pl_diagamma",
#         args=[pl.len()],
#         is_elementwise=True,
#     )

#     psi_n_neighbors = pl_plugin(
#         symbol="pl_diagamma",
#         args=[pl.lit(n_neighbors)],
#         is_elementwise=True,
#     )

#     return psi_n_samples + psi_n_neighbors - (psi_label_counts + psi_nb_cnt).mean()

#     return pl.max_horizontal(
#         pl.lit(0.0, dtype=pl.Float64),
#         psi_n_samples + psi_n_neighbors - psi_label_counts.mean() - psi_nb_cnt.mean(),
#     )



import polars as pl
import polars_ds as pds
import numpy as np
import scipy
import pytest
from polars_ds.linear_models import LR
from sklearn.linear_model import Lasso, LinearRegression, Ridge

SEED = 208

SIZE = 100_000
DF = (
    pds.frame(size=SIZE)
    .select(
        pds.random(0.0, 1.0).alias("x1"),
        pds.random(0.0, 1.0).alias("x2"),
        pds.random(0.0, 1.0).alias("x3"),
        pds.random(0.0, 1.0).alias("x4"),
        pds.random(0.0, 1.0).alias("x5"),
        pds.random_int(0, 4).alias("code"),
        pl.Series(name="id", values=range(SIZE)),
    )
    .with_columns(
        y=pl.col("x1") * 0.5
        + pl.col("x2") * 0.25
        - pl.col("x3") * 0.15
        + pl.col("x4") * 0.2
        - pl.col("x5") * 0.13
        + pds.random() * 0.0001,
    )
)


# Prepare data for Scikit-learn. We assume the Scikit-learn + Pandas combination.
# One can simply replace to_pandas() by to_numpy() to test the Scikit-learn + NumPy combination
PD_DF = DF.to_pandas()
X_VARS = ["x1", "x2", "x3", "x4", "x5"]
Y = ["y"]

SIZES = [1_000, 10_000, 50_000, 100_000]


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_matrix")
def test_pds_linear_regression_on_matrix(benchmark, n):
    df = DF.sample(n=n, seed=SEED)
    X = df.select(*X_VARS).to_numpy()
    y = df.select(*Y).to_numpy()

    @benchmark
    def func():
        model = LR()
        model.fit(X, y)


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_matrix")
def test_sklearn_linear_regression_on_matrix(benchmark, n):
    df = DF.sample(n=n, seed=SEED)
    X = df.select(*X_VARS).to_numpy()
    y = df.select(*Y).to_numpy()

    @benchmark
    def func():
        reg = LinearRegression(fit_intercept=False, n_jobs=-1)
        reg.fit(X, y)


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_matrix")
def test_numpy_linear_regression_on_matrix(benchmark, n):
    df = DF.sample(n=n, seed=SEED)
    X = df.select(*X_VARS).to_numpy()
    y = df.select(*Y).to_numpy()

    @benchmark
    def func():
        _ = np.linalg.lstsq(X, y)


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_matrix")
def test_scipy_linear_regression_on_matrix(benchmark, n):
    df = DF.sample(n=n, seed=SEED)
    X = df.select(*X_VARS).to_numpy()
    y = df.select(*Y).to_numpy()

    @benchmark
    def func():
        _ = scipy.linalg.lstsq(X, y, check_finite=False)


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_df")
def test_pds_linear_regression_on_df(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(
            pds.lin_reg(
                *X_VARS,
                target=Y[0],
            )
        )


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="linear_on_df")
def test_sklearn_linear_regression_on_df(benchmark, n):
    df = PD_DF.sample(n=n, random_state=SEED)

    @benchmark
    def func():
        reg = LinearRegression(fit_intercept=False, n_jobs=-1)
        reg.fit(df[X_VARS], df[Y])


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="lasso_on_df")
def test_pds_lasso_on_df(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.lin_reg(*X_VARS, target=Y[0], l1_reg=0.1))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="lasso_on_df")
def test_sklearn_lasso_on_df(benchmark, n):
    df = PD_DF.sample(n=n, random_state=SEED)

    @benchmark
    def func():
        reg = Lasso(alpha=0.1, fit_intercept=False)
        reg.fit(df[X_VARS], df[Y])


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="ridge_svd_on_df")
def test_ridge_svd_on_df(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.lin_reg(*X_VARS, target=Y[0], l2_reg=0.1, solver="svd"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="ridge_svd_on_df")
def test_sklearn_ridge_svd_on_df(benchmark, n):
    df = PD_DF.sample(n=n, random_state=SEED)

    @benchmark
    def func():
        reg = Ridge(alpha=0.1, fit_intercept=False, solver="svd")
        reg.fit(df[X_VARS], df[Y])


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="ridge_cholesky")
def test_ridge_cholesky_on_df(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.lin_reg(*X_VARS, target=Y[0], l2_reg=0.1, solver="cholesky"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="ridge_cholesky")
def test_sklearn_ridge_cholesky_on_df(benchmark, n):
    df = PD_DF.sample(n=n, random_state=SEED)

    @benchmark
    def func():
        reg = Ridge(alpha=0.1, fit_intercept=False, solver="cholesky")
        reg.fit(df[X_VARS], df[Y])



from __future__ import annotations


import polars_ds.exprs.expr_linear as pds_linear
import polars_ds.exprs.num as pds_num
import polars_ds.exprs.string as pds_str
import polars_ds.exprs.stats as pds_stats
import polars_ds.exprs.ts_features as pds_ts
import polars_ds.exprs.expr_knn as pds_knn
import polars_ds.exprs.metrics as pds_metrics

from polars_ds.compat import compat as pds2

# --- All functions can be Wrapped ---
# Test compatibility works (able to wrap the function for all the expression functions pds provides)


def test_lin_reg_works():
    # If this doesn't fail, the returned function is correct.
    # The rest of the usage depends on the user.
    for expr in pds_linear.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_metrics_works():
    for expr in pds_metrics.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_num_works():
    for expr in pds_num.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_str_works():
    for expr in pds_str.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_stats_works():
    for expr in pds_stats.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_ts_works():
    for expr in pds_ts.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True


def test_expr_knn_works():
    for expr in pds_knn.__all__:
        _ = getattr(pds2, expr)
    else:
        assert True



from __future__ import annotations

import polars as pl
import polars.selectors as cs
import warnings

# Typing
from collections.abc import Callable
from typing import List, Dict, Any

# Internal Dependencies
from polars_ds.typing import PolarsFrame


class PartitionHelper:
    """
    A transitory convenience class.
    """

    def __init__(
        self,
        df: PolarsFrame,
        by: str | List[str] | None,
        separator: str = "|",
        whole_df_name: str = "df",
    ):
        """
        Creates a Partition Result

        Parameters
        ----------
        df
            Either a Polars dataframe or a Lazyframe
        by
            Either None, or a string or a list of strings representing column names. If by
            is None, the entire df will be considered a partition.
        separator
            Separator for concatenating the names of different parts, if the partition is done
            by multiple columns
        whole_df_name
            If by is None, the name for the whole df.
        """
        if by is None:
            self.parts: Dict[str, pl.DataFrame] = {whole_df_name: df.lazy().collect()}
        else:
            cols = (
                df.select((cs.by_name(by)) & (cs.string() | cs.categorical() | cs.boolean()))
                .collect_schema()
                .names()
            )

            all_ok = cols[0] == by if isinstance(by, str) else sorted(cols) == sorted(by)
            if not all_ok:
                raise ValueError(
                    "Currently this only supports partitions by str, bool or categorical columns."
                )

            self.parts = {
                separator.join((str(k) for k in keys)): value
                for keys, value in df.lazy().collect().partition_by(by=by, as_dict=True).items()
            }

    def __repr__(self) -> str:
        output = ""
        for part, df in self.parts.items():
            output += f"Paritition: {part}\n"
            output += df.__repr__() + "\n"
        return output

    def head(self, n: int = 5) -> Dict[str, pl.DataFrame]:
        return {k: df.head(n) for k, df in self.parts.items()}

    def names(self) -> List[str]:
        return list(self.parts.keys())

    def get(self, part: str) -> pl.DataFrame | None:
        return self.parts.get(part, None)

    def apply(self, func: Callable[[str, pl.DataFrame], Any]) -> Dict[str, Any]:
        """
        Apply an arbitrary function to all parts in this partition.

        Parameters
        ----------
        func
            A function that takes in a str and a pl.DataFrame and outputs anything. The string
            represents the name of the segment. Note: this is usually a partial/lambda function with
            all other arguments provided.
        """
        output = {}
        for part, df in self.parts.items():
            try:
                output[part] = func(part, df)
            except Exception as e:
                warnings.warn(
                    f"An error occured while processing for the part: {part}. This partition is omitted.\nOriginal Error Message: {e}",
                    stacklevel=2,
                )

        return output



"""
Compatibility with other Dataframes.

This module provides compatibility with other dataframe libraries that:

1. Have a notion of Series
2. The Series implements the array protocal, which means it can be translated to NumPy array via
.__array__() method.

Since most dataframe libraries can turn their Series into NumPy (or vice versa) with 0 copy,
this compatibility layer has very little overhead. The only constraint is that the dataframe
must be eager, in the sense that data is already loaded in memory. The reason for this is that
the notion of a Series doesn't really exist in the lazy world, and lazy columns cannot be turned
to NumPy arrays.

When using this compatibility, the output is always a Polars Series. This is because the output
type could be Polars struct/list Series, which are Polars-specific types. It is up to the user
what to do with the output.

For example, in order to use PDS with Pandas dataframe, say df:pd.DataFrame, one needs to write

>>> from polars_ds.compat import compat as pds2
>>> # Output is a Polars Series.
>>> pds2.query_roc_auc(df_pd["actual"], df_pd["predicted"])
>>> # For more advanced queries
>>> pds2.lin_reg(
>>>     df_pd["x1"], df_pd["x2"], df_pd["x3"]
>>>     target = df_pd["y"],
>>>     return_pred = True
>>> )
"""

from ._compat import compat

import warnings

warnings.warn("The compatibility layer is considered experimental.", stacklevel=2)
